@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2012-2018 Snowflake Computing Inc. All right reserved.
+# Copyright (c) 2012-2019 Snowflake Computing Inc. All right reserved.
 #
 
 #
@@ -25,7 +25,8 @@ from .errors import (
     NotSupportedError, DataError, IntegrityError, ProgrammingError,
     OperationalError, InternalError)
 from .dbapi import (Timestamp, TimeFromTicks, Time, TimestampFromTicks, Date,
-                    DateFromTicks, DATETIME, ROWID, STRING, NUMBER, Json)
+                    DateFromTicks, Binary, Json,
+                    DATETIME, ROWID, STRING, NUMBER, BINARY)
 
 
 def Connect(**kwargs):
