@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2012-2018 Snowflake Computing Inc. All right reserved.
+# Copyright (c) 2012-2019 Snowflake Computing Inc. All right reserved.
 #
 import logging
 import os
@@ -13,6 +13,7 @@ from logging import getLogger
 from threading import Lock
 from time import strptime
 
+from .incident import IncidentAPI
 from . import errors
 from . import proxy
 from .auth import Auth
@@ -21,7 +22,10 @@ from .auth_keypair import AuthByKeyPair
 from .auth_oauth import AuthByOAuth
 from .auth_okta import AuthByOkta
 from .auth_webbrowser import AuthByWebBrowser
-from .chunk_downloader import SnowflakeChunkDownloader
+from .chunk_downloader import (
+    SnowflakeChunkDownloader,
+    DEFAULT_CLIENT_PREFETCH_THREADS,
+    MAX_CLIENT_PREFETCH_THREADS)
 from .compat import (
     TO_UNICODE, IS_OLD_PYTHON, urlencode, PY2, PY_ISSUE_23517, IS_WINDOWS)
 from .constants import (
@@ -29,11 +33,22 @@ from .constants import (
     PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY,
     PARAMETER_CLIENT_SESSION_KEEP_ALIVE,
     PARAMETER_CLIENT_TELEMETRY_ENABLED,
+    PARAMETER_CLIENT_TELEMETRY_OOB_ENABLED,
     PARAMETER_TIMEZONE,
     PARAMETER_SERVICE_NAME,
     PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL,
+    PARAMETER_CLIENT_PREFETCH_THREADS,
+    PARAMETER_CLIENT_VALIDATE_DEFAULT_PARAMETERS,
+    OCSPMode
+)
+from .cursor import SnowflakeCursor, LOG_MAX_QUERY_LENGTH
+from .description import (
+    SNOWFLAKE_CONNECTOR_VERSION,
+    PYTHON_VERSION,
+    PLATFORM,
+    CLIENT_NAME,
+    CLIENT_VERSION
 )
-from .cursor import SnowflakeCursor
 from .errorcode import (ER_CONNECTION_IS_CLOSED,
                         ER_NO_ACCOUNT_NAME, ER_OLD_PYTHON, ER_NO_USER,
                         ER_NO_PASSWORD, ER_INVALID_VALUE,
@@ -42,11 +57,6 @@ from .errorcode import (ER_CONNECTION_IS_CLOSED,
 from .errors import (Error, ProgrammingError, InterfaceError,
                      DatabaseError)
 from .network import (
-    SNOWFLAKE_CONNECTOR_VERSION,
-    PYTHON_VERSION,
-    PLATFORM,
-    CLIENT_NAME,
-    CLIENT_VERSION,
     DEFAULT_AUTHENTICATOR,
     EXTERNAL_BROWSER_AUTHENTICATOR,
     KEY_PAIR_AUTHENTICATOR,
@@ -57,10 +67,11 @@ from .network import (
 from .sqlstate import (SQLSTATE_CONNECTION_NOT_EXISTS,
                        SQLSTATE_FEATURE_NOT_SUPPORTED)
 from .telemetry import (TelemetryClient)
+from .telemetry_oob import TelemetryService
 from .time_util import (
     DEFAULT_MASTER_VALIDITY_IN_SECONDS,
     HeartBeatTimer, get_time_millis)
-from .util_text import split_statements, construct_hostname
+from .util_text import split_statements, construct_hostname, parse_account
 
 
 def DefaultConverterClass():
@@ -111,11 +122,13 @@ DEFAULT_CONFIGURATION = {
     u'internal_application_version': CLIENT_VERSION,
 
     u'insecure_mode': False,  # Error security fix requirement
+    u'ocsp_fail_open': True,  # fail open on ocsp issues, default true
     u'inject_client_pause': 0,  # snowflake internal
     u'session_parameters': None,  # snowflake session parameters
     u'autocommit': None,  # snowflake
     u'client_session_keep_alive': False,  # snowflake
     u'client_session_keep_alive_heartbeat_frequency': None,  # snowflake
+    u'client_prefetch_threads': 4,  # snowflake
     u'numpy': False,  # snowflake
     u'ocsp_response_cache_filename': None,  # snowflake internal
     u'converter_class': DefaultConverterClass(),
@@ -125,7 +138,10 @@ DEFAULT_CONFIGURATION = {
     u'paramstyle': None,  # standard/snowflake
     u'timezone': None,  # snowflake
     u'consent_cache_id_token': True,  # snowflake
-    u'service_name': None,  # snowflake
+    u'service_name': None,  # snowflake,
+    u'support_negative_year': True,  # snowflake
+    u'log_max_query_length': LOG_MAX_QUERY_LENGTH,  # snowflake
+    u'disable_request_pooling': False,  # snowflake
 }
 
 APPLICATION_RE = re.compile(r'[\w\d_]+')
@@ -147,6 +163,8 @@ class SnowflakeConnection(object):
     connect(..) to get the object.
     """
 
+    OCSP_ENV_LOCK = Lock()
+
     def __init__(self, **kwargs):
         self._lock_sequence_counter = Lock()
         self.sequence_counter = 0
@@ -169,10 +187,11 @@ class SnowflakeConnection(object):
         self.connect(**kwargs)
         self._telemetry = TelemetryClient(self._rest)
         self.telemetry_enabled = False
+        self.incident = IncidentAPI(self._rest)
 
     def __del__(self):
         try:
-            self.close()
+            self.close(retry=False)
         except:
             pass
 
@@ -185,6 +204,25 @@ class SnowflakeConnection(object):
         """
         return self._insecure_mode
 
+    @property
+    def ocsp_fail_open(self):
+        u"""
+        fail open mode. TLS cerificates continue to be validated. Revoked
+        certificates are blocked. Any other exceptions are disregarded.
+        :return:
+        """
+        return self._ocsp_fail_open
+
+    def _ocsp_mode(self):
+        """
+        OCSP mode. INSECURE, FAIL_OPEN or FAIL_CLOSED
+        :return:
+        """
+        if self.insecure_mode:
+            return OCSPMode.INSECURE
+        return OCSPMode.FAIL_OPEN \
+            if self.ocsp_fail_open else OCSPMode.FAIL_CLOSED
+
     @property
     def session_id(self):
         u"""
@@ -330,6 +368,22 @@ class SnowflakeConnection(object):
         self._client_session_keep_alive_heartbeat_frequency = value
         self._validate_client_session_keep_alive_heartbeat_frequency()
 
+    @property
+    def client_prefetch_threads(self):
+        u"""
+        Number of threads to download the result set
+        """
+        return self._client_prefetch_threads if \
+            self._client_prefetch_threads else DEFAULT_CLIENT_PREFETCH_THREADS
+
+    @client_prefetch_threads.setter
+    def client_prefetch_threads(self, value):
+        """
+        Number of threads to download the result set
+        """
+        self._client_prefetch_threads = value
+        self._validate_client_prefetch_threads()
+
     @property
     def rest(self):
         u"""
@@ -405,6 +459,18 @@ class SnowflakeConnection(object):
     def service_name(self, value):
         self._service_name = value
 
+    @property
+    def log_max_query_length(self):
+        return self._log_max_query_length
+
+    @property
+    def disable_request_pooling(self):
+        return self._disable_request_pooling
+
+    @disable_request_pooling.setter
+    def disable_request_pooling(self, value):
+        self._disable_request_pooling = True if value else False
+
     def connect(self, **kwargs):
         u"""
         Connects to the database
@@ -412,16 +478,18 @@ class SnowflakeConnection(object):
         logger.debug(u'connect')
         if len(kwargs) > 0:
             self.__config(**kwargs)
+            TelemetryService.get_instance().update_context(kwargs)
 
         self.__set_error_attributes()
         self.__open_connection()
 
-    def close(self):
+    def close(self, retry=True):
         u"""
         Closes the connection.
         """
         try:
             if not self.rest:
+                logger.debug(u'Rest object has been destroyed, cannot close session')
                 return
 
             # will hang if the application doesn't close the connection and
@@ -431,13 +499,14 @@ class SnowflakeConnection(object):
 
             # close telemetry first, since it needs rest to send remaining data
             logger.info('closed')
-            self._telemetry.close()
-            self.rest.delete_session()
+            self._telemetry.close(send_on_close=retry)
+            self.rest.delete_session(retry=retry)
             self.rest.close()
             self._rest = None
             del self.messages[:]
-        except:
-            pass
+            logger.debug(u'Session is closed')
+        except Exception as e:
+            logger.debug(u'Exception encountered in closing connection. ignoring...: %s', e)
 
     def is_closed(self):
         u"""
@@ -503,7 +572,8 @@ class SnowflakeConnection(object):
 
     def execute_string(self, sql_text,
                        remove_comments=False,
-                       return_cursors=True):
+                       return_cursors=True,
+                       **kwargs):
         """
         Executes a SQL text including multiple statements.
         This is a non-standard convenient method.
@@ -519,7 +589,7 @@ class SnowflakeConnection(object):
             cur = self.cursor()
             if return_cursors:
                 ret.append(cur)
-            cur.execute(sql, _is_put_get=is_put_or_get)
+            cur.execute(sql, _is_put_get=is_put_or_get, **kwargs)
         return ret
 
     def execute_stream(self, stream,
@@ -539,13 +609,24 @@ class SnowflakeConnection(object):
                   callable(getattr(errors, method))]:
             setattr(self, m, getattr(errors, m))
 
+    @staticmethod
+    def setup_ocsp_privatelink(app, hostname):
+        SnowflakeConnection.OCSP_ENV_LOCK.acquire()
+        ocsp_cache_server = \
+            u'http://ocsp.{}/ocsp_response_cache.json'.format(
+                hostname)
+        os.environ[
+            'SF_OCSP_RESPONSE_CACHE_SERVER_URL'] = ocsp_cache_server
+        logger.debug(u"OCSP Cache Server is updated: %s", ocsp_cache_server)
+        SnowflakeConnection.OCSP_ENV_LOCK.release()
+
     def __open_connection(self):
         u"""
         Opens a new network connection
         """
         self.converter = self._converter_class(
-            use_sfbinaryformat=False,
-            use_numpy=self._numpy)
+            use_numpy=self._numpy,
+            support_negative_year=self._support_negative_year)
 
         proxy.set_proxies(
             self.proxy_host, self.proxy_port, self.proxy_user,
@@ -561,12 +642,13 @@ class SnowflakeConnection(object):
                      self.host,
                      self.port)
 
+        if 'SF_OCSP_RESPONSE_CACHE_SERVER_URL' in os.environ:
+            logger.debug(
+                u"Custom OCSP Cache Server URL found in environment - %s",
+                os.environ['SF_OCSP_RESPONSE_CACHE_SERVER_URL'])
+
         if self.host.endswith(u".privatelink.snowflakecomputing.com"):
-            ocsp_cache_server = \
-                u'http://ocsp{}/ocsp_response_cache.json'.format(
-                    self.host[self.host.index('.'):])
-            os.environ['SF_OCSP_RESPONSE_CACHE_SERVER_URL'] = ocsp_cache_server
-            logger.debug(u"OCSP Cache Server is updated: %s", ocsp_cache_server)
+            SnowflakeConnection.setup_ocsp_privatelink(self.application, self.host)
         else:
             if 'SF_OCSP_RESPONSE_CACHE_SERVER_URL' in os.environ:
                 del os.environ['SF_OCSP_RESPONSE_CACHE_SERVER_URL']
@@ -593,6 +675,10 @@ class SnowflakeConnection(object):
         if self._timezone is not None:
             self._session_parameters[PARAMETER_TIMEZONE] = self._timezone
 
+        if self._validate_default_parameters:
+            # Snowflake will validate the requested database, schema, and warehouse
+            self._session_parameters[PARAMETER_CLIENT_VALIDATE_DEFAULT_PARAMETERS] = True
+
         if self.client_session_keep_alive:
             self._session_parameters[PARAMETER_CLIENT_SESSION_KEEP_ALIVE] = True
 
@@ -601,6 +687,10 @@ class SnowflakeConnection(object):
                 PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY] = \
                 self._validate_client_session_keep_alive_heartbeat_frequency()
 
+        if self.client_prefetch_threads:
+            self._session_parameters[PARAMETER_CLIENT_PREFETCH_THREADS] = \
+                self._validate_client_prefetch_threads()
+
         if self._authenticator == EXTERNAL_BROWSER_AUTHENTICATOR:
             # enable storing temporary credential in a file
             self._session_parameters[
@@ -659,6 +749,9 @@ class SnowflakeConnection(object):
             if u'protocol' not in kwargs:
                 setattr(self, u'_protocol', u'https')
 
+        if self._authenticator:
+            self._authenticator = self._authenticator.upper()
+
         if not self.user and self._authenticator != OAUTH_AUTHENTICATOR:
             # OAuth Authentication does not require a username
             Error.errorhandler_wrapper(
@@ -671,9 +764,6 @@ class SnowflakeConnection(object):
         if self._private_key:
             self._authenticator = KEY_PAIR_AUTHENTICATOR
 
-        if self._authenticator:
-            self._authenticator = self._authenticator.upper()
-
         if self._authenticator != EXTERNAL_BROWSER_AUTHENTICATOR and \
                 self._authenticator != OAUTH_AUTHENTICATOR and \
                 self._authenticator != KEY_PAIR_AUTHENTICATOR:
@@ -695,8 +785,16 @@ class SnowflakeConnection(object):
                     u'errno': ER_NO_ACCOUNT_NAME
                 })
         if u'.' in self._account:
-            # remove region subdomain
-            self._account = self._account[0:self._account.find(u'.')]
+            self._account = parse_account(self._account)
+
+        if self.ocsp_fail_open:
+            logger.info(
+                u'This connection is in OCSP Fail Open Mode. '
+                u'TLS Certificates would be checked for validity '
+                u'and revocation status. Any other Certificate '
+                u'Revocation related exceptions or OCSP Responder '
+                u'failures would be disregarded in favor of '
+                u'connectivity.')
 
         if self.insecure_mode:
             logger.info(
@@ -704,11 +802,12 @@ class SnowflakeConnection(object):
                 u'MEANS THE CERTIFICATE WILL BE VALIDATED BUT THE '
                 u'CERTIFICATE REVOCATION STATUS WILL NOT BE '
                 u'CHECKED.')
+
         elif self._protocol == u'https':
             if IS_OLD_PYTHON():
                 msg = (u"ERROR: The ssl package installed with your Python "
                        u"- version {0} - does not have the security fix. "
-                       u"Upgrade to Python 2.7.9/3.4.3 or higher.\n").format(
+                       u"Upgrade to Python 2.7.9/3.5.0 or higher.\n").format(
                     PYTHON_VERSION)
                 raise InterfaceError(
                     msg=msg,
@@ -742,9 +841,7 @@ class SnowflakeConnection(object):
         if logger.getEffectiveLevel() <= logging.DEBUG:
             logger.debug(
                 u'sql=[%s], sequence_id=[%s], is_file_transfer=[%s]',
-                u' '.join(
-                    line.strip() for line in
-                    data[u'sqlText'].split(u'\n')),
+                self._format_query_for_log(data[u'sqlText']),
                 data[u'sequenceId'],
                 is_file_transfer
             )
@@ -1016,6 +1113,15 @@ class SnowflakeConnection(object):
             self.client_session_keep_alive_heartbeat_frequency)
         return self.client_session_keep_alive_heartbeat_frequency
 
+    def _validate_client_prefetch_threads(self):
+        if self.client_prefetch_threads <= 0:
+            self._client_prefetch_threads = 1
+        elif self.client_prefetch_threads > MAX_CLIENT_PREFETCH_THREADS:
+            self._client_prefetch_threads = MAX_CLIENT_PREFETCH_THREADS
+        self._client_prefetch_threads = int(
+            self.client_prefetch_threads)
+        return self.client_prefetch_threads
+
     def _set_parameters(self, ret, session_parameters):
         """
         Set session parameters
@@ -1031,6 +1137,11 @@ class SnowflakeConnection(object):
             session_parameters[name] = value
             if PARAMETER_CLIENT_TELEMETRY_ENABLED == name:
                 self.telemetry_enabled = value
+            elif PARAMETER_CLIENT_TELEMETRY_OOB_ENABLED == name:
+                if value:
+                    TelemetryService.get_instance().enable()
+                else:
+                    TelemetryService.get_instance().disable()
             elif PARAMETER_CLIENT_SESSION_KEEP_ALIVE == name:
                 self.client_session_keep_alive = value
             elif PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY == \
@@ -1038,6 +1149,13 @@ class SnowflakeConnection(object):
                 self.client_session_keep_alive_heartbeat_frequency = value
             elif PARAMETER_SERVICE_NAME == name:
                 self.service_name = value
+            elif PARAMETER_CLIENT_PREFETCH_THREADS == name:
+                self.client_prefetch_threads = value
+
+    def _format_query_for_log(self, query):
+        ret = u' '.join(line.strip() for line in query.split(u'\n'))
+        return (ret if len(ret) < self.log_max_query_length
+                else ret[0:self.log_max_query_length] + '...')
 
     def __enter__(self):
         u"""
